:i	static char backup_sccs_id[] = "%W% %G%";
:1#include "V.h"
#include "sys_stat.h"
/*
* Here's the routine that "backs up" a file by renaming  it  according
* to  the convention in the BACKUP environment variable.  This is done
* recursively, of course, resulting in a series of files with more and
* more  final hyphens or whatever the convention is.  See backup.d for
* more details.
*
* This routine is in a separate file because it's a handy one to  make
* available for linking to other programs, posting on bulletin boards,
* and so on.
*
* The lines with ':' in column one are various sorts of debugging  and
* logging  hooks.   You  can  strip  them  out, and it should compile.
* Alternatively, you can get a copy of my V package, or  edit  these
* lines to fit whatever sort of audit/debug/log/trace facility you may
* have available.
*
* Note: if compiled with -DPROG, a main() will be generated that calls
* backup()  for  all  its  command-line arguments.  However, there are
* problems with some flakey C compilers (e.g.,  Sun's)  that  silently
* remove .o files when you compile directly to the executable; on such
* systems, it's better to make a program that links in these routines,
* and  test  it  via that program.  Otherwise, you'll find backup.o is
* mysteriously disappearing, and libV.a will always be remade.
*
* Note that the Backup() macro is defined to call backup(). The return
* value is zero for success, negative for failures. You may find errno
* set, since we make several system calls, but usually its value  will
* be meaningless.
*
* AUTHOR: John Chambers.
*   <jc%minya.uucp@trillian.mit.edu>
*   <jc%minya.uucp@harvard.edu>
*   <jc%minya.uucp@bu.edu>
*   <jc%minya.uucp@ora.com>
*   <jc%minya.uucp@bcs.org>
*   <jc%gtegsc3.sprint.com> [as of Jan 1994]
*/

#define BACKUPS 8 /* Default backup limit */
#define MAXBPSUFF 15	/* Longest legal backup suffix */

static int _bp();
global int  bpfailed = 0;	/* Number of removes that fail */
global Flag backupfl = 0;	/* Should we back up old files? */
global int  backups = -1;	/* Backup limit */
global char bpstyle = '-';	/* What kind of backup to do */
global int  bpsufnum = -1;	/* If positive, do numeric suffixes */
global char bpsuffix[MAXBPSUFF+1] = {0};	/* Suffix to append for backups */
static int  bpdepth =  0;	/* Current bpdepth of recursion */
global Flag preserve = 0;	/* Preserve file timestamps for cp/mv */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* CALL:
*   int backup(name)
*   	char *name;
*
* DESCRIPTION:
*
* This is the main user-callable routine for moving files  to  backup.
* All  we do here is check the backup option and maybe a bit of sanity
* checking, and then we call _bp(), which does the real work.
*
* AUTHOR: John Chambers.
*/
global int backup(name)
	char *name;
{	int   r=0;	/* Return value */
	char *p;

:f	Fenter("backup");
:3	V3 "Back up \"%s\"",name D;
	if (backups < 0) {	/* Have we determined the backup limit? */
		if (p = Getenv("BACKUP")) {	/* Look in the environment */
:5			V5 "BACKUP=\"%s\"",p D;
			bpopt(p);
		} else {
			backups = BACKUPS;
		}
:3		V3 "Limit: %d backups.",backups D;
	}
	errno = 0;
	r = _bp(name);	/* Recursive portion */
:f	Fexit;
	return r;
}

/*
* Some versions of V.h call these routines:
*/
global int d_backup(name) char *name; {return backup(name);}
global int d_bpopt(name) char *name; {return bpopt(name);}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* CALL:
*   int _bp(name)
*   	char *name;
*
* DESCRIPTION:
*
* Here's the recursive portion, which doesn't do any  error  checking.
* The  return  value  is minus the number of failures; 0 means that it
* all went OK.
*
* AUTHOR: John Chambers.
*/
static int _bp(name)
	char *name;
{	struct stat status;
	char *newname=0;
	char *dp, *sp, *p;
	int   e=0;	/* Errno value */
	int   c, n;
	int   r=0;	/* Return value */

:f	Fenter("_bp");
	if (++bpdepth > backups) {
:3		V3 "Max bpdepth %d reached.",backups D;
		Done;
	}
:2	V3 "back up \"%s\"",name D;
	if (Stat(name,&status) == 0) {	/* Does the file exist? */
:2		V4 "File \"%s\" exists...",name D;
		n = Strlen(name) + 2 + Strlen(bpsuffix);
		if (!(newname=(CP)MallocM(n,Fctname)))
			Fail;
		Switch(bpstyle) {
		case '*':
		case '.':
		case ';':
			if (bpsufnum > 0) {	/* Are we doing numeric suffixes? */
				for (dp = sp = p = name; c = *p; p++) {
					Switch(c) {
					case ';': case '*': case '-':
					case '.': if (c == bpstyle) dp = p; break;
					case '/': sp = p; break;
					}
				}
:5				V5 "dp: \"%s\"",dp D;
:5				V5 "sp: \"%s\"",sp D;
				if (dp > name && dp > sp) {	/* If there is a final ".x" */
:5					V5 "dp: \"%s\" is suffix",dp D;
					for (n=0, p=dp+1; c = *p; p++) {	/* Final number? */
						if (isdigit(c))
							n = (10 * n) + (c - '0');
						else break;			/* Found a non-digit */
					}
:5					V5 "p: \"%s\" n=%d",p,n D;
					if (c) {
:5						V5 "Non-num suffix: \"%s\"",dp D;
						while (*p) ++p;	/* Use entire name */
						n = 1;		/* and add ".1" */
					} else {
:5						V5 "Numeric suffix: %d \"%s\"",n,dp D;
						p = dp;		/* Use name up to the dot */
						++n;		/* Add one to the suffix */
					}
				} else {
					n = 1;
				}
				/*
				* Now p should point to the style char or the null,
				* and n should be the number to use as a suffix.
				*/
				Bcopy(name,newname,p-name);
				Sprintf(newname+(p-name),"%c%d",bpstyle,n);
				break;
			}
		case '-':
:5			V5 "style=%02X='%c' bpsuffix=\"%s\"",
:5				B8(bpstyle),Dsp(bpstyle),N(bpsuffix) D;
			Sprintf(newname,"%s%c%s",name,bpstyle,bpsuffix);
			break;
		case '#': {
			char *p, *q;
			int c0, c1;
			p = Lastfld('/',name);
:4			V4 "Last field \"%s\"",p D;
			if (p>name) {
				Strncpy(newname,name,p-name);
:4				V5 "New name: \"%s\"",newname D;
			}
			q = newname + (p-name);
			c0 = *p++;
			c1 = '#';
			*q = 0;
:6			V6 "q=%08X c0='%c' c1='%c' \"%s\"",q,c0,c1,newname D;
			while (c0) {
				*q++ = c1;
				c1 = c0;
				c0 = *p++;
:6				V6 "q=%08X c0='%c' c1='%c' \"%s\"",q,c0,c1,newname D;
				*q = 0;
:4				V5 "New name: \"%s\"",newname D;
			}
			*q++ = c1;
:6			V6 "q=%08X c0='%c' c1='%c' \"%s\"",q,c0,c1,newname D;
			*q = 0;
:4			V5 "New name: \"%s\"",newname D;
			break;
		}
		default:
:1			P1 "Can't handle %c%s backup yet.",bpstyle,bpsuffix D;
			Fail;
		}
:4		V4 "New name: \"%s\"",newname D;

		_bp(newname);	/**** Recursively rename it ****/

:4		V4 "rename \"%s\" as \"%s\"",name,newname D;
		if (Rename(name,newname) < 0) {
:2			V2 "%s: Can't rename \"%s\" as \"%s\" [Err %d=%s=%s]",
:2				pname,name,newname,Errinfo D;
			--r;
			if (errno != ENOENT) {
				++bpfailed;
			}
		}
:5	} else {
:5		V4 "File \"%s\" doesn't exist...",name D;
	}
done:
	Ifree(newname);
	--bpdepth;
:f	Fexit;
	return r;
fail:
	errno = ENOMEM;
	Ifree(newname);
	r = -1;
	Done;
}
#ifdef PROG
void main(ac,av)
	int   ac;
	char**av;
{	int   r=0;
	int   a;
:1	ac = Vinit(ac,av);
	for (a=1; a<ac; a++)
		r = backup(av[a]);
	exit(r);
}
#endif

/*
* This routine parses the BACKUP option string, which should be of
* the form:
*    <n><c><s>
* where
*    <n> is an optional limit to the number of levels.
*    <c> is a char ('-' or '.' or ';') as a separator.
*    <s> is an optional suffix to append.
* Actually, it's not a good idea to use <n> and <s> together.  The default
* that we were compiled with corresponds to "8-".
*/
global int bpopt(p)
	char *p;
{	int   r=0;
	int   i, l;
	char  c, d;
:f	Fenter("bpopt");
:6	V6 "Called for \"%s\"",NulC(p) D;
	if (!p) Fail;
	for (d = i = 0; (c = *p) && isdigit(c); p++) {
		i = (10 * i) + (c - '0');
		++d;
:6		V5 "DIGIT %d: c='%c' i=%d p=\"%s\"",d,c,i,p D;
	}
	l = Min(MAXBPSUFF,strlen(p));
:6	V5 "Count %d: c='%c' l=%d p=\"%s\"",d,c,l,p D;
	Switch(c = *p++) {	/* Test for a backup separator char */
	case '#':
		Bcopy(p,bpsuffix,l);	/* Save suffix and null */
		bpstyle = c;
		backups = d ? i : BACKUPS;
:3		V3 "Limit of %d backups for '%c' style, suffix=\"%s\".",
:3			backups,Dsp(bpstyle),bpsuffix D;
		break;
	case '*':
	case ';':
	case '-':
	case '.':	/* Check for a numeric suffix */
		Bcopy(p,bpsuffix,l);
		bpstyle = c;
		bpsufnum = -1;
		sscanf(bpsuffix,"%d",&bpsufnum);
:5		V5 "bpsufnum: %d bpsuffix=%s",bpsufnum,Dsps(bpsuffix,l) D;
		if (bpsufnum >= 0) {
			backups = bpsufnum;	/* The suffix is also the limit */
			bpsuffix[0] = 0;	/* We'll calculate the suffix */
		} else {
			backups = d ? i : 1;	/* Non-numeric suffix */
		}
:3		V3 "Limit of %d backups for '%c' style, suffix=\"%s\".",
:3			backups,Dsp(bpstyle),bpsuffix D;
		break;
:3	default:
:3		V2 "Unknown backup style '%c' in \"%s\"",Dsp(c),p D;
:3		break;
	}
done:
fail:
:5	V5 "Backup style='%c' backups=%d suffix=\"%s\".",bpstyle,backups,bpsuffix D;
:f	Fexit;
	return r;
}
